# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-10-02 07:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('app', '0003_auto_20170729_2305'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='tags',
            field=models.TextField(blank=True, null=True, verbose_name='tags'),
        ),
    ]
